/**
 * Copyright 2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 */
package com.sunlabs.archsim.fleet.ships.fetchAndIssue;

import java.util.Iterator;
import java.util.LinkedList;

import com.sunlabs.archsim.fleet.codeStorage.CodeBag;
import com.sunlabs.archsim.fleet.codeStorage.CodeStorage;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.fleet.common.OutputFunction;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

/**
 * @author ib27688
 * 
 */
class FetchComponent extends Component {
	
	Terminal in;
	Terminal ackIn;

	IssueComponent instructionIssueComponent;
	IssueComponent literalIssueComponent;
	
	FetchAction fetchAction;
	
	CodeStorage codeStorage= CodeStorage.getInstance(); 
	
	
	// note, we don't have parameters as an attribute.
	// we can remedy that by passing parameters to FetchAndIssue component
	public FetchComponent(String name, 
			IssueComponent instrComp, 
			IssueComponent litComp ) {
		super(name, null);
		
		buildFetchComponent(instrComp,litComp);
		
	} // FetchComponent constructor

	public void buildFetchComponent(IssueComponent instrComp, IssueComponent litComp) {
		instructionIssueComponent=  instrComp;
		literalIssueComponent= litComp;
		// assemble the component, start with the action
		fetchAction= new FetchAction(name+"FetchAction");
		addMember(fetchAction);
		// make the terminals
		in= new InputTerminal(FleetGlobal.IN, this);
		ackIn= new OutputTerminal(FleetGlobal.ACK_IN, this);
		// attach the terminals
		fetchAction.attach(in);
		fetchAction.attach(ackIn);	
	} // buildFetchComponent
	
	/**
	 * Access the terminal by name. Use either the base name or the base alias - 
	 * no prefix with names of the parents.
	 * @param terminal name
	 */
	public Terminal getTerminal(String name) {
		if (name.equals(FleetGlobal.IN) || name.equals(in.getBaseAlias())) {
			return in;
		}
		else if (name.equals(FleetGlobal.ACK_IN) || name.equals(ackIn.getBaseAlias())) {
			return ackIn;
		}
		else {
			return null;
		}
	}
	
	protected class FetchAction extends Action {
		
		OutputFunction fetchAckFn;
		
		public FetchAction(String name) {
			super(name);
			fetchAckFn= new FetchAckFn();
			outputFunctions.put(FleetGlobal.ACK_IN, fetchAckFn);
		}
		
		protected class FetchAckFn extends OutputFunction {
			
			public FetchAckFn() {
				super("FetchAckFn");
			} // FetchAckFn constructor
			
			/** Acknowledge receipt of a CBD
			 * Pass the instructions to the instructionComponent.
			 * Pass the literals to the literal component.
			 */
			public Data value() {
				String codeBagName=null;
				Object o= in.getData();
				boolean typeErrorDetected= true;
				if (o instanceof Payload) {
					Payload p= (Payload)o;
					if (p.data.value instanceof String) {
						codeBagName= (String)(p.data.value);
						typeErrorDetected= false;
					}
				}
				
				if (typeErrorDetected) {
					director.error(FetchAction.this.getName() 
							+ " received data of an inappropriate type."
							+ " Type= "
							+ o.getClass().getName()
							+ ", object = "
							+ o );
				}
				else if (codeBagName != null) {
					// get the codebag;
					CodeBag bag= codeStorage.getBag(codeBagName);
					Iterator<Payload> bagIterator= bag.iterator();
					// prepare the lists for instructions and for the literals 
					LinkedList<Payload> instrList= new LinkedList<Payload>();
					LinkedList<Payload> literalList= new LinkedList<Payload>();
					while (bagIterator.hasNext()) {
						Payload p= bagIterator.next();
						// does the instruction repeat?
						if (p.repeats ==1 || p.repeats < 0) {
							// classify the instruction
							if (p.type == Payload.PayloadType.INSTRUCTION) {
								instrList.add(p);
							}
							else if (p.type == Payload.PayloadType.LITERAL) {
								literalList.add(p);
							}
							else {
								// error
								director.error("Code bag "
										+ codeBagName
										+ " contains a Payload of unspecified type: "
										+ p);
							}
						}
						else if (p.repeats > 1) {
							// need to take care of cloning if data to make sure this always works
							// it will be OK for Integers and Strings, but if Data.value does not
							// implement a deep clone, two payloads may carry the same data object
							for (int i=0; i< p.repeats; i++) {
								if (p.type == Payload.PayloadType.INSTRUCTION) {
									instrList.add((Payload)p.clone());
								}
								else if (p.type == Payload.PayloadType.LITERAL) {
									literalList.add((Payload)p.clone());
								}
								else {
									// error
									director.error("Code bag "
											+ codeBagName
											+ " contains a Payload of unspecified type: "
											+ p);
								}	
							}
						} // else							
					} // while
					// payloads are classified, now pass them to appropriate 
					// componnents
					if (!instrList.isEmpty()) {
						instructionIssueComponent.addInstructions(instrList);
					}
					if (!literalList.isEmpty()) {
						literalIssueComponent.addInstructions(literalList);
					}
				} 
				// acknowledge receipt of the code bag				
				return new Data();
			}
			
		} //class FetchFn
		
	} // class FetchhAction

} // class Fetch
