/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

package com.sunlabs.archsim.fleet.ships.memory;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.BadDataException;
import com.sunlabs.archsim.fleet.common.CommonAckFunction;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.OutputFunction;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

/**
 * A simple model of memory.
 * Can read and write one data item at a time.
 * One read and one write can take place simultaeously.
 * If there is a conflict between an outstanding read and write,
 * then the completion of the write is not guaranteed to take place
 * before the completion of the read. That is, the read may either
 * return the data that was at the specified address before or after
 * the write.
 * 
 * @author ib27688
 *
 */

public class SimpleMemory extends Component {

	protected int[] memory;
	
	InputTerminal readAddressTerm;
	OutputTerminal ackReadAddressTerm;
	
	OutputTerminal readDataTerm;
	InputTerminal ackReadDataTerm;
	
	InputTerminal writeAddressTerm;
	OutputTerminal ackWriteAddressTerm;
	
	InputTerminal writeDataTerm;
	OutputTerminal ackWriteDataTerm;
	
	OutputTerminal writeNextTerm;
	InputTerminal ackWriteNextTerm;
	
	protected SimpleMemoryReadAction readAction;
	protected SimpleMemoryWriteAction writeAction;
	

	public SimpleMemory(String name, Parameters params) {
		super(name, null);	
		
		// process parameters
		Delay readDelayParam= null;
		Delay writeDelayParam= null;
		Delay ackDelayParam= null;
		Delay readAckDelayParam= null;
		Delay writeAckDelayParam= null;
		
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
				if (pName.equals(MemoryDefaults.SIMPLE_READ_DELAY)
						|| pName.equals(MemoryDefaults.READ_DELAY)) {
					try {
						int d= Integer.parseInt(sValue);
						readDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for "
								+ MemoryDefaults.SIMPLE_READ_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED.");						
					}					
				}
				else if (pName.equals(MemoryDefaults.SIMPLE_WRITE_DELAY)
						|| pName.equals(MemoryDefaults.WRITE_DELAY)) {
					try {
						int d= Integer.parseInt(sValue);
						writeDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for "
								+ MemoryDefaults.SIMPLE_WRITE_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED.");						
					}					
				}				
				else if (pName.equals(MemoryDefaults.ACK_DELAY)) {
					try {
						int d= Integer.parseInt(sValue);
						ackDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for "
								+ MemoryDefaults.ACK_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED.");						
					}					
				}
				else if (pName.equals(MemoryDefaults.ACK_READ_DELAY)) {
					try {
						int d= Integer.parseInt(sValue);
						readAckDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for "
								+ MemoryDefaults.ACK_READ_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED.");						
					}					
				}
				else if (pName.equals(MemoryDefaults.ACK_WRITE_DELAY)) {
					try {
						int d= Integer.parseInt(sValue);
						writeAckDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for "
								+ MemoryDefaults.ACK_WRITE_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED.");						
					}					
				}
				else {
					director.error("Component: " + name + ", id= " + getID() 
							+ ": Attribute " + pName + " with value " + sValue
							+ " is ignored");
				}
			} // for
		} // if		
		
		buildSimpleMemory();
		
		// apply the parameters if needed
		if (readDelayParam != null) {
			readAction.deliverDataFn.setDelay(readDelayParam);
		}
		if (writeDelayParam != null) {
			writeAction.writeFn.setDelay(writeDelayParam);
		}
		// set general delays first
		if (ackDelayParam != null) {
			readAction.readAckAddressFn.setDelay(readAckDelayParam);			
			writeAction.ackWriteDataFn.setDelay(writeAckDelayParam);
			writeAction.ackWriteDataFn.setDelay(writeAckDelayParam);
		}
		// overwrite general delays with specific ones
		if (readAckDelayParam != null) {
			readAction.readAckAddressFn.setDelay(readAckDelayParam);
		}
		if (writeAckDelayParam != null) {
			writeAction.ackWriteDataFn.setDelay(writeAckDelayParam);
			writeAction.ackWriteDataFn.setDelay(writeAckDelayParam);
		}
		
	} // SimpleMemoryConstructor
	
	
	public SimpleMemory(String name) {
		super(name, null);		
		buildSimpleMemory();
		
	} // SimpleMemoryConstructor
	
	protected void buildSimpleMemory() {
		// memory is represented as an array
		memory= new int[MemoryDefaults.MEMORY_SIZE_DEF];
		
		// make the action
		readAction= new SimpleMemoryReadAction("SimpleMemoryReadAction");
		addMember(readAction);
		writeAction= new SimpleMemoryWriteAction("SimpleMemoryWriteAction");
		addMember(writeAction);
		
		// now make the terminals, add them to the component, 
		// and attach them to the actions
		
		readAddressTerm= new InputTerminal(MemoryDefaults.READ_ADDRESS);
		addMember(readAddressTerm);
		readAction.attach(readAddressTerm);
		
		ackReadAddressTerm= new OutputTerminal(MemoryDefaults.ACK_READ_ADDRESS);
		addMember(ackReadAddressTerm);
		readAction.attach(ackReadAddressTerm);
		
		readDataTerm= new OutputTerminal(MemoryDefaults.READ_DATA);
		addMember(readDataTerm);
		readAction.attach(readDataTerm);
		
		ackReadDataTerm= new InputTerminal(MemoryDefaults.ACK_READ_DATA);
		addMember(ackReadDataTerm);
		readAction.attach(ackReadDataTerm);
		
		writeAddressTerm= new InputTerminal(MemoryDefaults.WRITE_ADDRESS);
		addMember(writeAddressTerm);
		writeAction.attach(writeAddressTerm);

		ackWriteAddressTerm= new OutputTerminal(MemoryDefaults.ACK_WRITE_ADDRESS);
		addMember(ackWriteAddressTerm);
		writeAction.attach(ackWriteAddressTerm);
		
		writeDataTerm= new InputTerminal(MemoryDefaults.WRITE_DATA);
		addMember(writeDataTerm);
		writeAction.attach(writeDataTerm);

		ackWriteDataTerm= new OutputTerminal(MemoryDefaults.ACK_WRITE_DATA);
		addMember(ackWriteDataTerm);
		writeAction.attach(ackWriteDataTerm);
		
		writeNextTerm= new OutputTerminal(MemoryDefaults.WRITE_NEXT);
		addMember(writeNextTerm);
		writeAction.attach(writeNextTerm);		

		ackWriteNextTerm= new InputTerminal(MemoryDefaults.ACK_WRITE_NEXT);
		addMember(ackWriteNextTerm);
		writeAction.attach(ackWriteNextTerm);				
	}
	
	/*
	 * Return a terminal by the base name. Return null if the name
	 * does not match any of the base terminal names. 
	 * @param name terminal name
	 */
	public Terminal getTerminal(String name) {
		if (name.equals(MemoryDefaults.READ_ADDRESS)) {
			return readAddressTerm;
		}
		else if (name.equals(MemoryDefaults.ACK_READ_ADDRESS)) {
			return ackReadAddressTerm;
		}		
		else if (name.equals(MemoryDefaults.READ_DATA)) {
			return readDataTerm;
		}
		else if (name.equals(MemoryDefaults.ACK_READ_DATA)) {
			return ackReadDataTerm;
		}
		else if (name.equals(MemoryDefaults.WRITE_ADDRESS)) {
			return writeAddressTerm;
		}
		else if (name.equals(MemoryDefaults.ACK_WRITE_ADDRESS)) {
			return ackWriteAddressTerm;
		}
		else if (name.equals(MemoryDefaults.WRITE_DATA)) {
			return writeDataTerm;
		}
		else if (name.equals(MemoryDefaults.ACK_WRITE_DATA)) {
			return ackWriteDataTerm;
		}
		else if (name.equals(MemoryDefaults.WRITE_NEXT)) {
			return writeNextTerm;
		}
		else if (name.equals(MemoryDefaults.ACK_WRITE_NEXT)) {
			return ackWriteNextTerm;
		}
		else 
			return null;
	} // getTerminal
	
	
	protected class SimpleMemoryReadAction extends Action {
		
		DeliverDataFn deliverDataFn;
		CommonAckFunction readAckAddressFn;
		
		public SimpleMemoryReadAction(String name) {
			super(name);
			
			deliverDataFn= new DeliverDataFn();
			readAckAddressFn= new CommonAckFunction();

			// get the ack delay
			Integer ad= globals.intValue(MemoryDefaults.SIMPLE_ACK_DELAY);
			Delay ackDelay= (ad != null)
				? new Delay(ad)
				: MemoryDefaults.SIMPLE_ACK_DELAY_DEF;
			// set the ack delay
			readAckAddressFn.setDelay(ackDelay);
			
			outputFunctions.put(MemoryDefaults.READ_DATA, deliverDataFn);
			outputFunctions.put(MemoryDefaults.ACK_READ_ADDRESS, readAckAddressFn);
		}
		
		class DeliverDataFn extends OutputFunction {
			
			public DeliverDataFn() {
				super("DeliverData");
				Integer fDelay= globals.intValue(MemoryDefaults.SIMPLE_READ_DELAY);
				functionDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: MemoryDefaults.SIMPLE_READ_DELAY_DEF;
			}
			
			public Data value() {
				
				if (readAddressTerm != null ) {
					// Get data - from the input
					Object pd= readAddressTerm.getData();
					// is this a payload object?
					if (pd instanceof Payload) {
						Data d= ((Payload)pd).data;
						//now inspect the value object
						if (d.value instanceof Integer) {
							int address = ((Integer)d.value).intValue();
							if (0<= address && address <= memory.length ) {
								return new Data(memory[address]);
							}
							else {
								// address out of bounds
								director.fatalError("SimpleMemoryReadAction " 
										+ getAlias() 
										+ ", id= " + getID() 
										+ ", attempted read with an out-of bounds address: "
										+ address);
								return new Data(false);
							}
						}
						else {
							// address not an integer
							director.fatalError("SimpleMemoryReadAction " 
									+ getAlias() 
									+ ", id= " + getID() 
									+ ", read requested with a non-integer address "
									+ d.value
									+ ", of type "
									+ d.value.getClass().getName());
							return new Data(false);			
						}
					}
					else {
						// input data not packaged as payload
						// address not an integer
						director.fatalError("SimpleMemoryReadAction " 
								+ getAlias() 
								+ ", id= " + getID() 
								+ ", received a non-payload object "
								+ pd
								+ ", of type "
								+ pd.getClass().getName());
						return new Data(false);						
					}
				}
				else {
					// no input terminal, report error
					director.fatalError("SimpleMemoryReadAction " 
							+ getAlias() 
							+ ", id= " + getID() 
							+ "does not have input terminal " 
							+ MemoryDefaults.READ_ADDRESS);						
					return new Data(false);

				}				
			}
			
		} // class DeliverDataFn
		
	} // class SimpleMemoryReadAction
	
	
	protected class SimpleMemoryWriteAction extends Action {
		
		AckWriteFn writeFn;
		CommonAckFunction ackWriteAddressFn;
		CommonAckFunction ackWriteDataFn;
		
		
		public SimpleMemoryWriteAction(String name) {
			super(name);
			
			writeFn= new AckWriteFn();
			ackWriteAddressFn= new CommonAckFunction();
			ackWriteDataFn= new CommonAckFunction();

			// get the ack delay
			Integer ad= globals.intValue(MemoryDefaults.SIMPLE_ACK_DELAY);
			Delay ackDelay= (ad != null)
				? new Delay(ad)
				: MemoryDefaults.SIMPLE_ACK_DELAY_DEF;
			// set the ack delay
			ackWriteAddressFn.setDelay(ackDelay);
			ackWriteDataFn.setDelay(ackDelay);
			
			outputFunctions.put(MemoryDefaults.WRITE_NEXT, writeFn);
			outputFunctions.put(MemoryDefaults.ACK_WRITE_ADDRESS, ackWriteAddressFn);
			outputFunctions.put(MemoryDefaults.ACK_WRITE_DATA, ackWriteDataFn);
			
		}
		
		private class AckWriteFn extends OutputFunction {
			
			public AckWriteFn() {
				super("AckWriteFn");
				Integer fDelay= globals.intValue(MemoryDefaults.SIMPLE_WRITE_DELAY);
				functionDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: MemoryDefaults.SIMPLE_WRITE_DELAY_DEF;
			} // AckWriteFn construnctor
			
			/**
			 * Get the data to be recorded into the memory
			 * @return the data to be recorded
			 * @throws BadDataException thrown when the data is not packaged as payload, 
			 * or is not an integer, or there is no input terminal for receiving data
			 */			
			protected int getAddress() throws BadDataException {
				int address=-1;
				String errorMessage=null;
				Object dataIn= null;
				boolean errorFound= false;
				
				if (writeAddressTerm != null ) {
					// Get data - from the input
					dataIn= writeAddressTerm.getData();
					// is this a payload object?
					if (dataIn instanceof Payload) {
						Data d= ((Payload)dataIn).data;
						//now inspect the value object
						if (d.value instanceof Integer) {
							address = ((Integer)d.value).intValue();
							if (0>= address && address > memory.length ) {
								// address out of bounds
								errorMessage= "SimpleMemoryWriteAction " 
										+ getAlias() 
										+ ", id= " + getID() 
										+ ", attempted to write with an out-of bounds address: "
										+ address;
								errorFound= true;
							}
						}
						else {
							// address not an integer
							errorMessage= "SimpleMemoryWriteAction " 
									+ getAlias() 
									+ ", id= " + getID() 
									+ ", write requested with a non-integer address: "
									+ d.value
									+ ", of type "
									+ d.value.getClass().getName();
							errorFound= true;		
						}
					}
					else {
						// input data not packaged as payload
						errorMessage= "SimpleMemoryWriteAction " 
								+ getAlias() 
								+ ", id= " + getID() 
								+ ", received a non-payload object for write address: "
								+ dataIn
								+ ", of type "
								+ dataIn.getClass().getName();
						errorFound= true;						
					}
				}
				else {
					// no input terminal, report error
					errorMessage= "SimpleMemoryWriteAction " 
							+ getAlias() 
							+ ", id= " + getID() 
							+ "does not have input terminal " 
							+ MemoryDefaults.WRITE_ADDRESS;						
					errorFound= true;
				}
				
				if (errorFound) {
					throw new BadDataException(errorMessage, dataIn);
				}
				
				return address;
				
			} // getAddress
			
			
			/**
			 * Get the data to be recorded into the memory
			 * @return the data to be recorded
			 * @throws BadDataException thrown when the data is not packaged as payload, 
			 * or is not an integer, or there is no input terminal for receiving data
			 */
			protected int getData() throws BadDataException {
				int data=-1;
				String errorMessage=null;
				Object dataIn= null;
				boolean errorFound= false;
				
				if (writeDataTerm != null ) {
					// Get data - from the input
					dataIn= writeDataTerm.getData();
					// is this a payload object?
					if (dataIn instanceof Payload) {
						Data d= ((Payload)dataIn).data;
						//now inspect the value object
						if (d.value instanceof Integer) {
							data = ((Integer)d.value).intValue();
						}
						else {
							// data not an integer
							errorMessage= "SimpleMemoryWriteAction " 
									+ getAlias() 
									+ ", id= " + getID() 
									+ ", write requested with a non-integer datum: "
									+ d.value
									+ ", of type "
									+ d.value.getClass().getName();
							errorFound= true;		
						}
					}
					else {
						// input data not packaged as payload
						errorMessage= "SimpleMemoryWriteAction " 
								+ getAlias() 
								+ ", id= " + getID() 
								+ ", received a non-payload object for write data: "
								+ dataIn
								+ ", of type "
								+ dataIn.getClass().getName();
						errorFound= true;						
					}
				}
				else {
					// no input terminal, report error
					errorMessage= "SimpleMemoryWriteAction " 
							+ getAlias() 
							+ ", id= " + getID() 
							+ "does not have input terminal " 
							+ MemoryDefaults.WRITE_DATA;						
					errorFound= true;
				}
				
				if (errorFound) {
					throw new BadDataException(errorMessage, dataIn);
				}
				
				return data;
				
			} // getData
			
			
			/** 
			 * Record the data in the memory and return the ack token.
			 */
			public Data value() {
				
				int address;
				int data;
				
				try {
					address= getAddress();
					data= getData();
				}
				catch (BadDataException e) {
					director.fatalError(e.getMessage());
					return new Data(false);
				}
				
				// no errors, record the data
				memory[address]= data;
				
				// return ack
				return new Data();				
				
			}
			
		} // class DeliverDataFn
	} // class SimpleMemoryWriteAction
	
} // SimpleMemory
